

/* Passing Grub's boot structures to the root task is not portable.
   Instead, we have our own protocol based on Grub's scheme: features
   are indicated via a flag variable and structures of sizes are
   passed inline to allow complete backwards compatibility.  All of
   the information is contiguous in memory.

   Consider the following example layout assuming four bytes words (a
   ~ implies a "do not care value"):

         struct boot_modules     +------------------+ ---.
  0x00           flags           | 3                |     \
  0x04           magic           | LADN             |      |
  0x08           size            | 0x64             |      | modules_offset
  0x0A           modules         | 2                |      |
  0x10           modules_offset  | 0x28             |      |
  0x14           cmdline_length  | 11               |      |
  0x18           cmdline_offset  | 0x1A             |      |
  0x1A           command line    | 'r' 'o' 'o' 't'  |      |
  0x20                           | '-' 's' 'e' 'r'  |      |
  0x24                           | 'v' 'r' 'r'  ~   |      |
         struct boot_module[0]   +------------------+     /
  0x28           flags           | 1                | <--' --.
  0x2A           size            | 0x20             |         \ 
  0x30           start           | 0x456000         |          |
  0x34           end             | 0x460000         |          | modules_offset
  0x38           cmdling_length  | 8                |          |      +
  0x3A           cmdling_offset  | 24 bytes         |          |    size
  0x40           command line    | 'p' 'a' 'g' 'e'  |          |
  0x44                           | 'r' ' ' '-' 'm'  |          |
         struct boot_module[1]   +------------------+         / 
  0x48           flags           | 1                | <------'
  0x4A           size            | 8                |
  0x50           start           | 0x500000         |
  0x54           length          | 0x510A00         |
  0x58           cmdline_length  | 8                |
  0x5A           cmdline_offset  | 24 bytes         |
  0x60           command line    | 'f' 'o' 'o' ' '  |
  0x64                           | '-' 'b' 'a' 'r'  |
                                 +------------------+

  If a have flag is not set, the corresponding data structures will
  still be present, however, they will contain "don't care" values.

  'struct boot_module's will always start on a word boundary.
 */

#ifndef __BOOT_MODULES__
#define __BOOT_MODULES__

#include <l4/arch/types.h>

#ifdef L4_BIG_ENDIAN
#define BOOT_MODULE_MAGIC ('L' + ('A' << 8) + ('D' << 16) + ('N' << 24))
#else
#define BOOT_MODULE_MAGIC (('L' << 24) + ('A' << 16) + ('D' << 8) + 'N')
#endif /* L4_BIG_ENDIAN  */

struct boot_module
{
  struct
  {
    L4_Word_t have_commandline: 1; /* Have a command line.  */
    L4_Word_t padding:sizeof (L4_Word_t) * 8 - 1;
  } flags;

  /* Always valid.  */
  L4_Word_t size;                /* Size of this structure, plus
				    commandline, et al.  Must end on a
				    word boundary.  */

  L4_Word_t start;               /* Start address (physical).  */
  L4_Word_t end;                 /* End address (i.e. last valid byte).  */

  /* Valid if have_commandline is true.  */
  L4_Word_t commandline_length;  /* Length of command line (bytes).  */
  L4_Word_t commandline_offset;  /* Offset from start of this structure.  */

  /* Offset in structure is not fixed.  */
  char commandline[0];           /* Not necessarily nul terminated.  */
};

struct boot_modules_header
{
  struct
  {
    L4_Word_t have_modules: 1;
    L4_Word_t have_commandline: 1;
    L4_Word_t padding: sizeof (L4_Word_t) * 8 - 2;
  } flags;

  /* Always valid.  */
  char magic[4];

  L4_Word_t size; /* Size of this strucure and all of the modules.  */

  /* Valid if have_modules is true.  */
  L4_Word_t module_count; /* Number of modules.  */
  L4_Word_t modules_offset; /* Offset of the first module from the
			       beginning for this structure; must be a
			       word boundary.  */

  /* Valid if have_commandline is true.  */
  L4_Word_t commandline_length;  /* Length of command line (bytes).  */
  L4_Word_t commandline_offset;  /* Offset from start of this structure.  */

  /* Offset is in structure is not fixed.  */
  char commandline[0];           /* Not necessarily nul terminated.  */
};

#endif /* __BOOT_MODULES__  */
