/* 
   Copyright (C) 1995, 1996, 2002 Free Software Foundation, Inc.
   Written by Michael I. Bushnell.

   This file is part of the GNU Hurd.

   The GNU Hurd is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   The GNU Hurd is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "server-handles.h"
#include "sh-internal.h"
#include <pthread.h>

#define INHIBITED (SERVER_HANDLES_INHIBITED | SERVER_HANDLES_INHIBIT_WAIT)

error_t
server_handle_begin_rpc (void *server_handle_info, server_handle_t sh,
			 msg_id_t msg_id, struct rpc_info *info)
{
  int *block_flags = 0;

  struct server_handle_info *shi = server_handle_info;
  
  pthread_mutex_lock (&_server_handle_lock);
  
  do
    {
      /* Is the RPC permitted? */
      if (! _server_handle_valid_rpc (shi, sh))
	{
	  pthread_mutex_unlock (&_server_handle_lock);
	  return EOPNOTSUPP;
	}
  
      if (_server_handle_flags & INHIBITED)
	/* All RPC's are inhibited.  */
	block_flags = &_server_handle_flags;
      else if (shi->bucket->flags & INHIBITED)
	/* RPC's are inhibited for this server_handle's bucket.  */
	block_flags = &shi->bucket->flags;
      else if (shi->class->flags & INHIBITED)
	/* RPC's are inhibited for this server_handle's class.  */
	block_flags = &shi->class->flags;
      else if (shi->flags & INHIBITED)
	/* RPC's are inhibited for this server_handle itself.  */
	block_flags = &shi->flags;
      else
	block_flags = 0;

      if (block_flags)
	/* We maybe want to block.  */
	{
	  if (msg_id)
	    /* See if this particular RPC shouldn't be inhibitable.  */
	    {
	      struct server_handle_msg_id_range *range
		= shi->class->uninhibitable_rpcs;
	      while (range)
		if (msg_id >= range->start && msg_id < range->end)
		  {
		    block_flags = 0;
		    break;
		  }
		else
		  range = range->next;
	    }

	  if (block_flags)
	    {
	      *block_flags |= SERVER_HANDLES_BLOCKED;

	      /* If we are canceled and the only one blocking,
		 SERVER_HANDLES_BLOCKED will still be turned on, but
		 that's ok as it will just cause a (harmless) extra
		 pthread_cond_broadcast().  */
	      pthread_cleanup_push ((void (*) (void *)) pthread_mutex_unlock,
				    (void *) &_server_handle_lock);
	      pthread_cond_wait (&_server_handle_block,
				 &_server_handle_lock);
	      pthread_cleanup_pop (0);
	    }
	}
    }
  while (block_flags);
  
  /* Record that that an RPC is in progress */
  info->thread = pthread_self ();
  info->notifies = 0;

  info->next = shi->current_rpcs;
  if (shi->current_rpcs)
    shi->current_rpcs->prevp = &info->next;
  info->prevp = &shi->current_rpcs;
  shi->current_rpcs = info;

  shi->class->rpcs++;
  shi->bucket->rpcs++;
  _server_handle_total_rpcs++;

  pthread_mutex_unlock (&_server_handle_lock);

  return 0;
}
