/* 
   Copyright (C) 1996,2001,02 Free Software Foundation, Inc.
   Written by Michael I. Bushnell, p/BSG.

   This file is part of the GNU Hurd.

   The GNU Hurd is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   The GNU Hurd is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA. */

#include "server-handles.h"
#include "sh-internal.h"

#include <assert.h>
#include <pthread.h>

/* Create and return in RESULT a new server_handle in CLASS and BUCKET; SIZE
   bytes will be allocated to hold the server_handle structure and whatever
   private data the user desires.  If INSTALL is set, put the server_handle
   right into BUCKET's server_handle set.  */
error_t
_server_handle_create_internal (struct server_handle_class *class, 
				struct server_handle_bucket *bucket,
				size_t size, void *result,
				int install)
{
  error_t err;
  struct server_handle_info *sh;

  void cleanup (void *x)
    {
      pthread_mutex_unlock (&_server_handle_lock);
      _server_handle_info_fini (sh);
      free (sh);
    }

  if (size < sizeof (struct server_handle_info))
    size = sizeof (struct server_handle_info);

  sh = malloc (size);
  if (! sh)
    return ENOMEM;

  sh->refcnt = 1;
  sh->weakrefcnt = 0;
  sh->cancel_threshold = 0;
  sh->flags = 0;
  sh->current_rpcs = 0;
  sh->class = class;
  sh->bucket = bucket;

  err = _server_handle_info_init (sh);
  if (err)
    {
      free (sh);
      return err;
    }
  
  pthread_mutex_lock (&_server_handle_lock);

  while (1)
    {
      if (class->flags & SERVER_HANDLE_CLASS_NO_ALLOC)
	{ 
	  class->flags |= SERVER_HANDLE_CLASS_ALLOC_WAIT;

	  pthread_cleanup_push (cleanup, 0);
	  pthread_cond_wait (&_server_handle_block,
			     &_server_handle_lock);
	  pthread_cleanup_pop (0);
	}
      else if (bucket->flags & SERVER_HANDLE_BUCKET_NO_ALLOC)
	{
	  bucket->flags |= SERVER_HANDLE_BUCKET_ALLOC_WAIT;

	  pthread_cleanup_push (cleanup, 0);
	  pthread_cond_wait (&_server_handle_block,
			     &_server_handle_lock);
	  pthread_cleanup_pop (0);
	}
      else
	break;
    }

  if (install)
    {
      err = _server_handle_info_install (sh);
      if (err)
	{
	  cleanup (0);
	  return err;
	}
    }

  sh->next = class->server_handle;
  sh->prevp = &class->server_handle;
  if (class->server_handle)
    class->server_handle->prevp = &sh->next;
  class->server_handle = sh;

  bucket->count++;
  class->count++;

  pthread_mutex_unlock (&_server_handle_lock);

  *(void **)result = sh;
  return 0;
}

