/* Create a new server_handle structure using an externally supplied receive right

   Copyright (C) 1995, 1996 Free Software Foundation, Inc.

   Written by Michael I. Bushnell.

   This file is part of the GNU Hurd.

   The GNU Hurd is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   The GNU Hurd is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "server-handles.h"
#include <assert.h>
#include <pthread.h>
#include <hurd/ihash.h>
#include <mach/notify.h>

/* For an existing receive right SERVER_HANDLE, create and return in
   RESULT a new server_handle structure; BUCKET, SIZE, and CLASS args
   are as for server_handle_create_server_handle. */
error_t
server_handle_import_server_handle (struct server_handle_class *class,
				     struct server_handle_bucket *bucket,
				     server_handle_t server_handle,
				     size_t size, void *result)
{
  error_t err;
  server_handle_status_t stat;
  struct server_handle_info *pi;
  server_handle_t foo;

#warning lots of evil mach operations
  err = server_handle_get_receive_status (mach_task_self (), server_handle, &stat);
  if (err)
    return err;

  if (size < sizeof (struct server_handle_info))
    size = sizeof (struct server_handle_info);
  
  pi = malloc (size);
  if (! pi)
    return ENOMEM;
  
  pi->class = class;
  pi->refcnt = 1 + !!stat.mps_srights;
  pi->weakrefcnt = 0;
  pi->cancel_threshold = 0;
  pi->mscount = stat.mps_mscount;
  pi->flags = stat.mps_srights ? SERVER_HANDLE_HAS_SENDRIGHTS : 0;
  pi->server_handle_right = server_handle;
  pi->current_rpcs = 0;
  pi->bucket = bucket;
  
  pthread_mutex_lock (&_server_handle_lock);
  
 loop:
  if (class->flags & SERVER_HANDLE_CLASS_NO_ALLOC)
    { 
      class->flags |= SERVER_HANDLE_CLASS_ALLOC_WAIT;
      if (pthread_cond_wait (&_server_handle_block,
			     &_server_handle_lock))
	goto cancelled;
      goto loop;
    }
  if (bucket->flags & SERVER_HANDLE_BUCKET_NO_ALLOC)
    {
      bucket->flags |= SERVER_HANDLE_BUCKET_ALLOC_WAIT;
      if (pthread_cond_wait (&_server_handle_block,
			     &_server_handle_lock))
	goto cancelled;
      goto loop;
    }

  err = ihash_add (bucket->htable, server_handle, pi, &pi->hentry);
  if (err)
    goto lose;

  pi->next = class->server_handle;
  pi->prevp = &class->server_handle;
  if (class->server_handle)
    class->server_handle->prevp = &pi->next;
  class->server_handle = pi;
  bucket->count++;
  class->count++;
  pthread_mutex_unlock (&_server_handle_lock);
  
#ifdef MACH
  port_move_member (mach_task_self (), server_handle, bucket->portset);
#endif

  if (stat.mps_srights)
    {
      err = server_handle_request_notification (mach_task_self (), server_handle,
					    MACH_NOTIFY_NO_SENDERS, 
					    stat.mps_mscount, 
					    server_handle, MACH_MSG_TYPE_MAKE_SEND_ONCE,
					    &foo);
      assert_perror (err);
      if (foo != MACH_SERVER_HANDLE_NULL)
	server_handle_deallocate (mach_task_self (), foo);
    }

  *(void **)result = pi;
  return 0;

 cancelled:
  err = EINTR;
 lose:
  pthread_mutex_unlock (&_server_handle_lock);
  free (pi);

  return err;
}
