/*
   Copyright (C) 1995,96,97, 2002 Free Software Foundation, Inc.
   Written by Michael I. Bushnell.

   This file is part of the GNU Hurd.

   The GNU Hurd is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   The GNU Hurd is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "server-handles.h"
#include <spin-lock.h>
#include <assert.h>
#include <pthread.h>
#include <mach/message.h>

void
server_handle_manage_multithread (struct server_handle_bucket *bucket,
				  server_handle_demuxer_type demuxer,
				  int thread_timeout,
				  int global_timeout,
				  void (*hook)())
{
  volatile int nreqthreads;
  volatile int totalthreads;
  pthread_spinlock_t lock = PTHREAD_SPINLOCK_INITIALIZER;

  auto int thread_function (int);

  int
  internal_demuxer (msg_header_t *inp,
		    msg_header_t *outheadp)
    {
      int spawn = 0;
      int status;
      struct server_handle_info *sh;
      struct rpc_info link;
      register mig_reply_header_t *outp = (mig_reply_header_t *) outheadp;
      static const msg_type_t RetCodeType = {
		/* msgt_name = */		MACH_MSG_TYPE_INTEGER_32,
		/* msgt_size = */		32,
		/* msgt_number = */		1,
		/* msgt_inline = */		TRUE,
		/* msgt_longform = */		FALSE,
		/* msgt_deallocate = */		FALSE,
		/* msgt_unused = */		0
	};

      pthread_spin_lock (&lock);
      assert (nreqthreads);
      nreqthreads--;
      if (nreqthreads == 0)
	spawn = 1;

      if (spawn)
	{
	  pthread_t tid;

	  totalthreads++;
	  nreqthreads++;
	  pthread_spin_unlock (&lock);

	  err = pthread_create (&tid, 0, thread_function, 0);
	  assert_perror (err);
	  err = pthread_detach (tid);
	  assert_perror (err);
	}
      else
	pthread_spin_unlock (&lock);

      /* Fill in default response. */
      outp->Head.msgh_bits 
	= MACH_MSGH_BITS(MACH_MSGH_BITS_REMOTE(inp->msgh_bits), 0);
      outp->Head.msgh_size = sizeof *outp;
      outp->Head.msgh_remote_port = inp->msgh_remote_port;
      outp->Head.msgh_local_port = MACH_PORT_NULL;
      outp->Head.msgh_seqno = 0;
      outp->Head.msgh_id = inp->msgh_id + 100;
      outp->RetCodeType = RetCodeType;
      outp->RetCode = MIG_BAD_ID;

      sh = server_handle_lookup (bucket, inp->msgh_local_server_handle, 0);
      if (sh)
	{
	  error_t err = server_handle_begin_rpc (sh, 0, inp->msgh_id, &link);
	  if (err)
	    {
	      outp->RetCode = err;
	      status = 1;
	    }
	  else
	    {
	      pthread_mutex_lock (&_server_handle_lock);
	      if (inp->msgh_seqno < sh->cancel_threshold)
		hurd_thread_cancel (link.thread);
	      pthread_mutex_unlock (&_server_handle_lock);
	      status = demuxer (inp, outheadp);
	      server_handle_end_rpc (sh, &link);
	    }
	  server_handle_deref (sh);
	}
      else
	{
	  outp->RetCode = EOPNOTSUPP;
	  status = 1;
	}

      pthread_spin_lock (&lock);
      nreqthreads++;
      pthread_spin_unlock (&lock);

      return status;
    }

  int
  thread_function (int master)
    {
      int timeout;
      error_t err;

      if (hook)
	(*hook) ();

      if (master)
	timeout = global_timeout;
      else
	timeout = thread_timeout;

    startover:

      do
	err = msg_server_timeout (internal_demuxer, 0, bucket->portset,
				  timeout ? MACH_RCV_TIMEOUT : 0,
				  timeout);
      while (err != MACH_RCV_TIMED_OUT);

      if (master)
	{
	  pthread_spin_lock (&lock);
	  if (totalthreads != 1)
	    {
	      pthread_spin_unlock (&lock);
	      goto startover;
	    }
	}
      else
	{
	  pthread_spin_lock (&lock);
	  nreqthreads--;
	  totalthreads--;
	  pthread_spin_unlock (&lock);
	}
      return 0;
    }

  thread_timeout = global_timeout = 0; /* XXX */

  nreqthreads = 1;
  totalthreads = 1;
  thread_function (1);
}





