/* Object handle library for server construction
   Copyright (C) 1993,94,95,96,97,99,2000,02 Free Software Foundation, Inc.
   Written by Michael I. Bushnell.

   This file is part of the GNU Hurd.

   The GNU Hurd is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   The GNU Hurd is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef _HURD_SERVER_HANDLES_
#define _HURD_SERVER_HANDLES_

#include <stdlib.h>
#include <pthread.h>
#include <hurd.h>

#include <hurd/bits/server-handles.h>

/* These are global values for common flags used in the various structures.
   Not all of these are meaningful in all flag fields.  */
/* Block RPC's.  */
#define SERVER_HANDLES_INHIBITED	0x0100
/* If INHIBITED, someone is blocked.  */
#define SERVER_HANDLES_BLOCKED		0x0200
/* Someone wants to start inhibit.  */
#define SERVER_HANDLES_INHIBIT_WAIT	0x0400
/* Block allocation. */
#define SERVER_HANDLES_NO_ALLOC		0x0800
/* Someone wants to allocate.  */
#define SERVER_HANDLES_ALLOC_WAIT	0x1000

struct server_handle_info
{
  SYSDEPS_SERVER_HANDLE_INFO

  int refcnt;
  int weakrefcnt;

  /* Initialized to zero and only set by interrupt_rpc.  When
     receiving messages, any message with an ipc which has a sequence
     number less than this value is discarded.  */
  msg_seqno_t cancel_threshold;

  int flags;

  /* List of current rpcs sent to this handle.  */
  struct rpc_info *current_rpcs;

  struct server_handle_bucket *bucket;
  struct server_handle_class *class;

  void **hentry;

  /* Links on server_handle_class list.  */
  struct server_handle_info *next, **prevp;
};
/* FLAGS above are the following: */
#define SERVER_HANDLE_HAS_USERS		0x0001
#define SERVER_HANDLE_INHIBITED		SERVER_HANDLES_INHIBITED
#define SERVER_HANDLE_BLOCKED		SERVER_HANDLES_BLOCKED
#define SERVER_HANDLE_INHIBIT_WAIT	SERVER_HANDLES_INHIBIT_WAIT

struct server_handle_bucket
{
  SYSDEPS_SERVER_HANDLE_BUCKET

  /* Number of exant rpcs; incremented in begin_rpc and decremented in
     end_rpc.  */
  int rpcs;

  int flags;

  /* Number of user handles in this bucket.  */
  int count;

  /* Hash from handle ids to server_handle_info structures.  */
  struct ihash *htable;

  struct server_handle_bucket *next;
};
/* FLAGS above are the following: */
#define SERVER_HANDLE_BUCKET_INHIBITED		SERVER_HANDLES_INHIBITED
#define SERVER_HANDLE_BUCKET_BLOCKED		SERVER_HANDLES_BLOCKED
#define SERVER_HANDLE_BUCKET_INHIBIT_WAIT	SERVER_HANDLES_INHIBIT_WAIT
#define SERVER_HANDLE_BUCKET_NO_ALLOC		SERVER_HANDLES_NO_ALLOC
#define SERVER_HANDLE_BUCKET_ALLOC_WAIT		SERVER_HANDLES_ALLOC_WAIT

struct server_handle_class
{
  SYSDEPS_SERVER_HANDLE_CLASS

  /* Number of exant rpcs.  */
  int rpcs;

  int flags;

  /* Count of user handles in this class.  */
  int count;

  /* List of handles on this class.  */
  struct server_handle_info *server_handle;

  void (*clean_routine) (void *);
  void (*dropweak_routine) (void *);

  struct server_handle_msg_id_range *uninhibitable_rpcs;
};
/* FLAGS are the following: */
#define SERVER_HANDLE_CLASS_INHIBITED		SERVER_HANDLES_INHIBITED
#define SERVER_HANDLE_CLASS_BLOCKED		SERVER_HANDLES_BLOCKED
#define SERVER_HANDLE_CLASS_INHIBIT_WAIT	SERVER_HANDLES_INHIBIT_WAIT
#define SERVER_HANDLE_CLASS_NO_ALLOC		SERVER_HANDLES_NO_ALLOC
#define SERVER_HANDLE_CLASS_ALLOC_WAIT		SERVER_HANDLES_ALLOC_WAIT

struct rpc_info
{
  pthread_t thread;
  struct rpc_info *next, **prevp;
  struct rpc_notify *notifies;
  struct rpc_info *interrupted_next;
};

/* An rpc has requested interruption on a server_handle notification.  */
struct rpc_notify
{
  struct rpc_info *rpc;		/* Which rpc this is for. */
  struct server_handle_notify *notify; /* Which server_handle/request this refers too. */

  struct rpc_notify *next;	/* Notify for this rpc. */
  unsigned pending;		/* Number of requests this represents.  */

  struct rpc_notify *next_req;	/* rpc for this notify.  */
  struct rpc_notify **prev_req_p; /* who points to this rpc_notify. */
};

/* A notification request on a (not necessarily registered) user handle.  */
struct server_handle_notify
{
  user_handle_t handle;		/*  */
  msg_id_t what;		/* MACH_NOTIFY_* */
  unsigned pending : 1;		/* There's a notification outstanding.  */
  pthread_mutex_t lock;

  struct rpc_notify *reqs;	/* Which rpcs are notified by this server_handle. */
  struct server_handle_notify *next, **prevp; /* Linked list of all notified server_handle.  */
};

/* A linked list of server_handle that have had notification requested.  */
extern struct server_handle_notify *_server_handle_notifications;

/* Free lists for notify structures.  */
extern struct server_handle_notify *_server_handle_free_server_handle_notifies;
extern struct rpc_notify *_server_handle_free_rpc_notifies;

/* Remove RPC from the list of notified rpcs, cancelling any pending
   notifications.  _SERVER_HANDLES_LOCK should be held.  */
void _server_handle_remove_notified_rpc (struct rpc_info *rpc);

struct server_handle_msg_id_range
{
  msg_id_t start, end;
  struct server_handle_msg_id_range *next;
};

/* This is the initial value for the uninhibitable_rpcs field in new
   server_handle_class structures.  The user may define this variable;
   the default value contains only an entry for
   interrupt_operation.  */
extern struct server_handle_msg_id_range *
  server_handle_default_uninhibitable_rpcs;

/* Server handle creation and frobbing */

/* Create and return a new bucket. */
struct server_handle_bucket *server_handle_create_bucket (void);

/* Create and return a new server_handle class.  If nonzero,
   CLEAN_ROUTINE will be called for each allocated server_handle
   object in this class when it is being destroyed.  If nonzero,
   DROPWEAK_ROUTINE will be called to request weak references to be
   dropped.  (If DROPWEAK_ROUTINE is null, then normal references and
   hard references will be identical for server_handle of this
   class.)  */
struct server_handle_class *
server_handle_create_class (void (*clean_routine) (void *),
			     void (*dropweak_routine) (void *));

/* Create and return in RESULT a new server_handle in CLASS and
   BUCKET; SIZE bytes will be allocated to hold the server_handle
   structure and whatever private data the user desires.  */
error_t server_handle_create (struct server_handle_class *class,
			      struct server_handle_bucket *bucket,
			      size_t size, void *result);

/* Just like server_handle_create_server_handle, except don't
   actually put the server_handle into the portset underlying BUCKET.
   This is intended to be used for cases where the server_handle right
   must be given out before the server_handle is fully initialized;
   with this call you are guaranteed that no RPC service will occur on
   the server_handle until you have finished initializing it and
   installed it into the portset yourself. */
error_t server_handle_create_noinstall (struct server_handle_class *class,
					struct server_handle_bucket *bucket,
					size_t size, void *result);

/* Finishes the creation of the handle started by
   server_handle_create_server_handle_noinstall.  */
error_t server_handle_install (void *server_handle);


/* Destroy the handle currently associated with SERVER_HANDLE and
   allocate a new one.  On failure, the handle is in an undefined
   state and should be destroyed.  */
error_t server_handle_reallocate (void *server_handle);

/* Destroy SERVER_HANDLE.  Always returns 0, for convenient use as an
   iterator.  */
error_t server_handle_destroy (void *server_handle);

/* Return a new server handle associated with SERVER_HANDLE for task
   TARGET_TASK.  */
user_handle_t server_handle_create_user_handle (void *server_handle,
						task_t target_task);


/* Reference counting */

/* Look up HANDLE and return the associated server_handle structure,
   allocating a reference.  If the call fails, return 0.  If BUCKET is
   nonzero, then it specifies a bucket to search; otherwise all
   buckets will be searched.  If CLASS is nonzero, then the lookup
   will fail if HANDLE is not in CLASS. */
void *server_handle_lookup (struct server_handle_bucket *bucket,
			    server_handle_t handle,
			    struct server_handle_class *class);

/* Allocate another reference to SERVER_HANDLE. */
void server_handle_ref (void *server_handle);

/* Allocate a weak reference to SERVER_HANDLE. */
void server_handle_ref_weak (void *server_handle);

/* Drop a reference to SERVER_HANDLE. */
void server_handle_deref (void *server_handle);

/* Drop a weak reference to SERVER_HANDLE. */
void server_handle_deref_weak (void *server_handle);

#if 0
/* NEAL: No sender notifications and deadname notifications are
   different under L4.  Here we will use task death to determine when
   there are no more senders (or that the server (i.e. via dead name
   notification) has died.  How should we abstract this?  */

/* The user is responsible for listening for no senders notifications;
   when one arrives, call this routine for the SERVER_HANDLE the
   message was sent to, providing the MSCOUNT from the
   notification. */
void server_handle_no_senders (void *server_handle,
			       server_handle_mscount_t mscount);

/* NEAL: Not used anywhere.  */
void server_handle_dead_name (void *notify,
			      server_handle_t dead_name);
#endif

/* Block server_handle creation of new server_handle in CLASS.
   Return the number of server_handle currently in CLASS. */
int server_handle_count_class (struct server_handle_class *class);

/* Block server_handle creation of new server_handle in BUCKET.
   Return the number of server_handle currently in BUCKET. */
int server_handle_count_bucket (struct server_handle_bucket *bucket);

/* Permit suspended server_handle creation (blocked by
   server_handle_count_class) to continue. */
void server_handle_enable_class (struct server_handle_class *class);

/* Permit suspend server_handle creation (blocked by
   server_handle_count_bucket) to continue. */
void server_handle_enable_bucket (struct server_handle_bucket *bucket);

/* Call FUN once for each server_handle in BUCKET. */
error_t server_handle_bucket_iterate (struct server_handle_bucket *bucket,
				       error_t (*fun)(void *server_handle));

/* Call FUN once for each server_handle in CLASS. */
error_t server_handle_class_iterate (struct server_handle_class *class,
				      error_t (*fun)(void *server_handle));

/* Internal entrypoint for both server_handle_bucket_iterate and
   server_handle_class_iterate.  If CLASS is non-null, call FUN only
   for server_handle in that class.  */
error_t _server_handle_bucket_class_iterate (struct server_handle_bucket
					     *bucket,
					     struct server_handle_class
					     *class,
					     error_t (*fun)(void *),
					     int have_server_handle_lock);

/* RPC management */

/* Call this when an RPC is beginning on SERVER_HANDLE.  INFO should
   be allocated by the caller and will be used to hold dynamic state.
   If this RPC should be abandoned, return EDIED; otherwise we return
   zero. */
error_t server_handle_begin_rpc (void *server_handle,
				 server_handle_t h,
				 msg_id_t msg_id,
				 struct rpc_info *info);

/* Call this when an RPC is concluding.  Args must be as for the
   paired call to server_handle_begin_rpc. */
void server_handle_end_rpc (void *server_handle, struct rpc_info *info);

/* Begin handling operations for the server_handle in BUCKET, calling
   DEMUXER for each incoming message.  Return if TIMEOUT is nonzero
   and no messages have been received for TIMEOUT milliseconds.  Use
   only one thread (the calling thread).  */
void server_handle_manage_one_thread (struct server_handle_bucket *bucket,
				      server_handle_demuxer_type demuxer,
				      int timeout);

/* Begin handling operations for the server_handle in BUCKET, calling
   DEMUXER for each incoming message.  Return if GLOBAL_TIMEOUT is
   nonzero and no messages have been receieved for GLOBAL_TIMEOUT
   milliseconds.  Create threads as necessary to handle incoming
   messages so that no server_handle is starved because of
   sluggishness on another server_handle.  If LOCAL_TIMEOUT is
   non-zero, then individual threads will die off if they handle no
   incoming messages for LOCAL_TIMEOUT milliseconds.  HOOK (if not
   null) will be called in each new thread immediately after it is
   created. */
void server_handle_manage_multithread (struct server_handle_bucket *bucket,
					server_handle_demuxer_type demuxer,
					int thread_timeout,
					int global_timeout,
					void (*hook)(void));

/* Interrupt any pending RPC on SERVER_HANDLE.  Wait for all pending
   RPC's to finish, and then block any new RPC's starting on that
   server_handle. */
error_t server_handle_inhibit_rpcs (void *server_handle);

/* Similar to server_handle_inhibit_server_handle_rpcs, but affects
   all server_handle in CLASS. */
error_t server_handle_inhibit_class_rpcs (struct server_handle_class *class);

/* Similar to server_handle_inhibit_server_handle_rpcs, but affects
   all server_handle in BUCKET. */
error_t server_handle_inhibit_bucket_rpcs (struct server_handle_bucket *bucket);

/* Similar to server_handle_inhibit_server_handle_rpcs, but affects
   all server_handle whatsoever. */
error_t server_handle_inhibit_all_rpcs (void);

/* Reverse the effect of a previous
   server_handle_inhibit_server_handle_rpcs for this SERVER_HANDLE,
   allowing blocked RPC's to continue. */
void server_handle_resume_rpcs (void *server_handle);

/* Reverse the effect of a previous server_handle_inhibit_class_rpcs
   for CLASS. */
void server_handle_resume_class_rpcs (struct server_handle_class *class);

/* Reverse the effect of a previous server_handle_inhibit_bucket_rpcs
   for BUCKET. */
void server_handle_resume_bucket_rpcs (struct server_handle_bucket *bucket);

/* Reverse the effect of a previous server_handle_inhibit_all_rpcs. */
void server_handle_resume_all_rpcs (void);

/* Cancel (with thread_cancel) any RPC's in progress on
   SERVER_HANDLE. */
void server_handle_interrupt_rpcs (void *server_handle);

/* If the current thread's rpc has been interrupted with
   server_handle_interrupt_rpcs, return true (and clear the
   interrupted flag).  */
int server_handle_self_interrupted ();

/* Add RPC to the list of rpcs that have been interrupted.  */
void _server_handle_record_interruption (struct rpc_info *rpc);

/* NEAL: How are we going to do notifications?  */
#if 0
/* Arrange for hurd_cancel to be called on RPC's thread if OBJECT gets
   notified that any of the things in COND have happened to
   SERVER_HANDLE.  RPC should be an rpc on OBJECT.  */
error_t
server_handle_interrupt_rpc_on_notification (void *object,
					      struct rpc_info *rpc,
					      server_handle_t server_handle,
					      msg_id_t what);

/* Arrange for hurd_cancel to be called on the current thread, which
   should be an rpc on OBJECT, if SERVER_HANDLE gets notified with the
   condition WHAT.  */
error_t
server_handle_interrupt_self_on_notification (void *object,
					       server_handle_t server_handle,
					       msg_id_t what);

/* NEAL: We do not have this (see ports_no_senders for explanation).  */

/* Some handy aliases.  */
#define server_handle_interrupt_self_on_server_handle_death(obj,
							     server_handle) \
  server_handle_interrupt_self_on_notification (obj, server_handle, \
						 MACH_NOTIFY_DEAD_NAME)

/* Interrupt any rpcs on OBJECT that have requested such.  */
void server_handle_interrupt_notified_rpcs (void *object,
					     server_handle_t server_handle,
					     msg_id_t what);
#endif

/* Default servers */

/* NEAL: This will not work at all.  We must rethink this.  */
#if 0

/* A notification server that calls the
   server_handle_do_mach_notify_* routines.  */
int server_handle_notify_server (msg_header_t *, msg_header_t *);

/* Notification server routines called by server_handle_notify_server.  */
extern kern_return_t server_handle_do_mach_notify_dead_name (server_handle_t
							      notify,
							      server_handle_t
							      deadserver_handle);
extern kern_return_t server_handle_do_mach_notify_msg_accepted (server_handle_t notify,
								 server_handle_t name);
extern kern_return_t server_handle_do_mach_notify_no_senders (server_handle_t server_handle,
							       server_handle_mscount_t
							       count);
extern kern_return_t server_handle_do_mach_notify_server_handle_deleted (server_handle_t notify, server_handle_t name);
extern kern_return_t server_handle_do_mach_notify_server_handle_destroyed (server_handle_t
									    notify,
									    server_handle_t
									    name);
extern kern_return_t server_handle_do_mach_notify_send_once (server_handle_t
							      notify);

#endif

/* Private data */
extern pthread_mutex_t _server_handle_lock;
extern pthread_cond_t _server_handle_block;
extern struct server_handle_bucket *_server_handle_all_buckets;
extern int _server_handle_total_rpcs;
extern int _server_handle_flags;
#define _SERVER_HANDLES_INHIBITED	SERVER_HANDLES_INHIBITED
#define _SERVER_HANDLES_BLOCKED		SERVER_HANDLES_BLOCKED
#define _SERVER_HANDLES_INHIBIT_WAIT	SERVER_HANDLES_INHIBIT_WAIT

/* _server_handle_lock must be held on entry and will be released on
   exit.  */
void _server_handle_complete_deallocate (struct server_handle_info *);
error_t
_server_handle_create_internal (struct server_handle_class *,
				struct server_handle_bucket *,
				size_t, void *, int);

#endif
