/* Transfer the receive right from one server_handle structure to another
   Copyright (C) 1996 Free Software Foundation, Inc.
   Written by Michael I. Bushnell, p/BSG.

   This file is part of the GNU Hurd.

   The GNU Hurd is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   The GNU Hurd is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA. */


#include "server-handles.h"
#include <assert.h>
#include <hurd/ihash.h>

error_t
server_handle_transfer_right (void *tostruct, 
			       void *fromstruct)
{
  struct server_handle_info *topi = tostruct;
  struct server_handle_info *frompi = fromstruct;
  server_handle_t server_handle;
  int dereffrompi = 0;
  int dereftopi = 0;
  int hassendrights = 0;
  error_t err;

  pthread_mutex_lock (&_server_handle_lock);

  /* Fetch the server_handle in FROMPI and clear its use */
  server_handle = frompi->server_handle_right;
  if (server_handle != MACH_PORT_NULL)
    {
      ihash_locp_remove (frompi->bucket->htable, frompi->hentry);
      frompi->server_handle_right = MACH_PORT_NULL;
      if (frompi->flags & SERVER_HANDLE_HAS_SENDRIGHTS)
	{
	  frompi->flags &= ~SERVER_HANDLE_HAS_SENDRIGHTS;
	  hassendrights = 1;
	  dereffrompi = 1;
	}
    }
  
  /* Destroy the existing right in TOPI. */
  if (topi->server_handle_right != MACH_PORT_NULL)
    {
      ihash_locp_remove (topi->bucket->htable, topi->hentry);
      err = mach_mod_refs (mach_task_self (), topi->server_handle_right,
			   MACH_PORT_RIGHT_RECEIVE, -1);
      assert_perror (err);
      if ((topi->flags & SERVER_HANDLE_HAS_SENDRIGHTS) && !hassendrights)
	{
	  dereftopi = 1;
	  topi->flags &= ~SERVER_HANDLE_HAS_SENDRIGHTS;
	}
      else if (((topi->flags & SERVER_HANDLE_HAS_SENDRIGHTS) == 0) && hassendrights)
	{
	  topi->flags |= SERVER_HANDLE_HAS_SENDRIGHTS;
	  topi->refcnt++;
	}
    }
  
  /* Install the new right in TOPI. */
  topi->server_handle_right = server_handle;
  topi->cancel_threshold = frompi->cancel_threshold;
  topi->mscount = frompi->mscount;
  
  if (server_handle)
    {
      ihash_add (topi->bucket->htable, server_handle, topi, &topi->hentry);
      if (topi->bucket != frompi->bucket)
        {
	  err = mach_move_member (mach_task_self (), server_handle,
				  topi->bucket->server_handleet);
	  assert_perror (err);
	}
    }
  
  pthread_mutex_unlock (&_server_handle_lock);
  
  /* Take care of any lowered reference counts. */
  if (dereffrompi)
    server_handle_server_handle_deref (frompi);
  if (dereftopi)
    server_handle_server_handle_deref (topi);
  return 0;
}


  
