#include <zalloc.h>
#include <assert.h>
#include <machine/vmparam.h>

#ifdef DEBUG
#include <stdio.h>
#endif

/* Zalloc: A fast zone allocator.  This is not a general purpose
   allocator.  If you attempt to use it as such, you will find that it
   is very inefficient.  It is, however, designed to be very fast and
   to be used as a base for building a more general purpose allocator.

   Memory is kept in zones.  Zones are of sizes 2 ** N and all memory
   is aligned on a similar boundary.  Typically, the smallest zone
   will be the system page size.  Memory of any size can be added to
   the pool as long as it is a multiple of the smallest zone: it is
   broken up as necessary.

   Memory can be added to the pool by calling the zfree function with
   the address of the buffer and its size.  The buffer is broken up as
   a function of its alignment and size using the buddy system (as
   described by e.g. Knuth).  Consider the following: zfree (4k, 16k).
   This says that a buffer of size 16k starting at address 4k should
   be added to the system.  Although the size of the buffer is a power
   of 2 (2 ** 14 = 16k), it cannot be added to the 16k zone: it has
   the wrong alignment.  Instead, the initial 4k are broken off, added
   to the 4k zone, the next 8k to the 8k zone and the final 4k to the
   4k zone.  If, as memory is added to a zone, its buddy is present,
   the two buffers are buddied up and promoted to the next zone.  For
   instance, if the 4k buffer at address 20k was present during the
   previous zfree, the bufer at 16k would have been combined with this
   and the new larger buffer would have been added to the 8k zone.

   When allocating memory, the smallest zone that is larger than or
   equal to the desired size is selected.  If the zone is exhausted,
   the allocator will look in the next larger zone and break up a
   buffer to satisfy the request.  This continues recursively if
   necessary.  If the desired size is smaller than the buffer that is
   selected, the difference is returned to the system.  For instance,
   if an allocation request of 12k is made, the system will start
   looking in the 16k zone.  If it finds that that zone is exhausted,
   it will select a buffer from the 32k zone and place the top half in
   the 16k zone and use the lower half for the allocation.  However,
   as this is 4k too much, the extra is returned to the 4k zone.

   When making allocations, the system will not look for adjacent
   memory blocks: if an allocation request of e.g. 8k is issued and
   there is no memory in the 8k zones and above, the 4k zone will not
   be searched for false buddies.  That is, if in the 4k zone there is
   a buffer starting at 4k and 8k, the allocator will make no effort
   to search for them.  Note that they could not have been combined
   during the zfree as 4k's buddy is at 0k and 8k's buddy is at
   12k.  */


/* A free block list ordered by address.  Blocks are of size 2 ** N
   and aligned on a similar boundary.  Since the contents of a block
   does not matter (it is free), the block itself contains this
   structure at its start address.  */
struct block
{
  struct block *next;
  struct block *prev;
};

/* The zones.  */
struct block *zone[ZONES] = { 0, };

#ifdef DEBUG
#define DODEBUG(level, func) \
	do { if ((level) <= DEBUG) { func; } } while (0)
#else
#define DODEBUG(level, func) ((void) (0))
#endif

/* Add the block VBLOCK of size SIZE to the pool of available
   memory.  */
void
zfree (void *vblock, L4_Word_t size)
{
  int zn;
  struct block *block = vblock;

  DODEBUG (1, printf ("%s: freeing %d bytes starting at %x.\n",
		      __FUNCTION__, size, block));

  {
    int i;
    for (i = 0; i < size / sizeof (int *); i ++)
      ((int *) block)[i] = 0x00FEED00;
  }

  if (size < PAGE_SIZE)
    enter_kdebug ("zfree: attempt to allocate less than a page.");
  
  if (size & (PAGE_SIZE - 1))
    enter_kdebug ("zfree: allocation size is not page aligned.");

  /* All blocks must be stored with 2 ** N alignment.  Pick off the
     smaller aligned blocks (i.e. at the start of BLOCK) working
     towards blocks of larger alignment and then back down.  Consider
     a block starting at address 4k of size 16kb.  */
  do
    {
      DODEBUG (2, printf ("%s: %d bytes to go.\n", __FUNCTION__, size));
      for (zn = 0;; zn ++)
	{
	  int z;

	  DODEBUG (2, printf ("%s: block = %x; zn = %d (0x%x)\n",
			      __FUNCTION__, block, zn, ZONE_SIZE (zn)));

	  if (size >= ZONE_SIZE (zn + 1)
	      && zn < ZONES
	      && ! ((L4_Word_t) block & (1 << (zn + PAGE_SHIFT))))
	    continue;

	  DODEBUG (2, printf ("%s: Releasing block %x of %d bytes.\n",
			      __FUNCTION__, block, ZONE_SIZE (zn)));

	  /* We now have the largest 2 ** N aligned block possible
	     starting at address BLOCK of ZONE_SIZE(ZN).  Add the
	     block to the system.  Buddy up if possible.  */
	  z = zn;

	  while (1)
	    {
	      struct block *l, *r;

	      for (l = 0, r = zone[z]; r && block > r; l = r, r = r->next)
		{
		  assert (! l || l->next == r);
		  assert (! r || l == r->prev);
		  assert (l < block);
		}

	      DODEBUG (2, printf ("%s: l = %x; block = %x; r = %x\n",
				  __FUNCTION__, l, block, r));

	      if (z < ZONES)
		{
		  DODEBUG (2, printf ("%s: Trying to coalesce.\n",
				      __FUNCTION__));

		  if (l && ((L4_Word_t) l
			    == ((L4_Word_t) block & ~(1 << (z + PAGE_SHIFT)))))
		    /* Buddy on left.  */
		    {
		      assert ((void *) l + ZONE_SIZE (z) == block);
		      DODEBUG (2, printf ("%s: Have a left buddy at %x!\n",
					  __FUNCTION__, l));
		      DODEBUG (2, enter_kdebug ("left buddy"));

		      if (l->prev)
			l->prev->next = l->next;
		      else
			zone[z] = l->next;
		      if (l->next)
			l->next->prev = l->prev;

		      block = l;
		      z ++;
		      continue;
		    }
		  if (r && ((L4_Word_t) block
			    == ((L4_Word_t) r & ~(1 << (z + PAGE_SHIFT)))))
		    /* Buddy on right.  */
		    {
		      assert ((void *) block + ZONE_SIZE (z) == r);
		      DODEBUG (2, printf ("%s: Have a right buddy at %x!\n",
					  __FUNCTION__, r));
		      DODEBUG (2, enter_kdebug ("right buddy"));

		      if (r->prev)
			r->prev->next = r->next;
		      else
			zone[z] = r->next;
		      if (r->next)
			r->next->prev = r->prev;

		      z ++;
		      continue;
		    }
		}

	      DODEBUG (2, printf ("%s: Inserting %x of size %d (0x%x).\n",
				  __FUNCTION__, block, ZONE_SIZE (z),
				  ZONE_SIZE (z)));

	      /* Could not coalesce.  Just insert.  */
	      block->next = r;
	      if (block->next)
		block->next->prev = block;
	      block->prev = l;

	      if (block->prev)
		block->prev->next = block;
	      else
		zone[z] = block;
		  
	      break;
	    }

	  /* ZONE_SIZE (Z), not ZN -- BLOCK potentially moved left and
	     grew as we buddied up.  */
	  block = (void *) block + ZONE_SIZE (z);

	  size -= ZONE_SIZE (zn);
	  DODEBUG (2, printf ("%s: Block now at %x, %d (0x%x) bytes to go.\n",
			      __FUNCTION__, block, size, size));
	  if (size < ZONE_SIZE (zn))
	    break;
	}
    }
  while (size > 0);

  DODEBUG (1, printf ("%s: leaving.\n", __FUNCTION__));
  DODEBUG (2, zalloc_dump_zones (__FUNCTION__));
  DODEBUG (3, enter_kdebug ("End of zfree"));
}

/* Allocate a block of memory of size SIZE.  Size *must* be a multiple
   of the system page size. */
void *
zalloc (L4_Word_t size)
{
  int z;
  void *b;

  DODEBUG (1, printf ("%s: request for %d (0x%x) bytes.\n",
		      __FUNCTION__, size, size));

  if (size < PAGE_SIZE)
    enter_kdebug ("zalloc: attempt to allocate less than a page.");
  
  if (size & (PAGE_SIZE - 1))
    enter_kdebug ("zalloc: allocation size is not page aligned.");

  if (size > ZONE_SIZE (ZONES - 1))
    return 0;

  /* Determine minimum zone size.  */
  for (z = 0; ZONE_SIZE (z) < size && z < ZONES; z ++)
    ;

  /* Find the smallest zone which fits the request and has memory
     available.  */
  for (; z < ZONES && ! zone[z]; z ++)
    ;

  b = zone[z];
  if (! b)
    return 0;

  DODEBUG (2, printf ("%s: allocating block of %d (0x%x) bytes.\n",
		      __FUNCTION__, ZONE_SIZE (z), ZONE_SIZE (z)));

  zone[z] = ((struct block *) b)->next;
  if (zone[z])
    zone[z]->prev = 0;

  if (ZONE_SIZE (z) > size)
    {
      DODEBUG (1,
	       printf ("%s: Trimming %d (0x%x) bytes from the tail of %x\n",
		       __FUNCTION__, ZONE_SIZE (z) - size,
		       ZONE_SIZE (z) - size, b));
      zfree (b + size, ZONE_SIZE (z) - size);
    }

#if 0
  {
    int i;
    for (i = 0; i < size / sizeof (int *); i ++)
      ((int *) b)[i] = 0xAA00FF00;
  }
#endif

  DODEBUG (1, printf ("%s: allocated block 0x%x.\n", __FUNCTION__, b));
  DODEBUG (2, zalloc_dump_zones (__FUNCTION__));
  DODEBUG (3, enter_kdebug ("End of zalloc"));

  return b;
}

/* Dump the internal data structures.  */
#ifdef DEBUG
void
zalloc_dump_zones (char *prefix)
{
  int i;
  struct block *b;
  int available = 0;
  int printempty = 0;

  for (i = ZONES - 1; i >= 0; i --)
    if (zone[i] || printempty)
      {
	printempty = 1;
	printf ("%s: %x: {", prefix, ZONE_SIZE (i));
	for (b = zone[i]; b; b = b->next)
	  {
	    assert (! b->prev || b->prev->next == b);
	    assert (! b->next || b->next->prev == b);
	    assert (b->prev < b);
	    assert (! b->next || b < b->next);
	    available += ZONE_SIZE (i);
	    printf ("%x%s", b, (b->next ? ", " : ""));
	  }
	printf ("}\n");
      }

  printf ("%s: %d (0x%x) bytes availables.\n", prefix, available, available);
}
#endif
