/* parse.c - Boot command lines parser.
   Copyright (C) 2002 Free Software Foundation, Inc.
   Written by Wolfgang Jhrling

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.  */

#include <stdio.h>  /* printf () and NULL */
#include <stdlib.h> /* malloc (), realloc () and free () */
#include <string.h> /* memcpy (), strlen (), strchr () and strstr () */
#include <assert.h> /* assert () */
#include <hurd/debug.h>

#include "parse.h"

/* Like malloc (), but only return when successful.  */
static void *
xmalloc (int size)
{
  void *ptr = malloc (size);

  assert (ptr != NULL);
  return ptr;
}

/* Like realloc (), but only return when successful.  */
static void *
xrealloc (void *buf, int size)
{
  void *ptr = realloc (buf, size);

  assert (ptr != NULL);
  return ptr;
}



/* Linked list of variables.  */
static struct variable
{
  struct variable *next;
  char *name;
  char *value;
} *vars;

/* See `parse.h' for a description.  */
void
parser_cleanup (void)
{
  struct variable *ptr, *tmp;

  DODEBUG (3, printf ("Entered %s ()\n", __FUNCTION__));

  for (ptr = vars; ptr != NULL; ptr = tmp)
    {
      free (ptr->name);
      free (ptr->value);
      tmp = ptr->next;
      free (ptr);
    }
}

/* See `parse.h' for a description.  */
int
add_variable (const char *name, char *value)
{
  struct variable *new = xmalloc (sizeof (struct variable));
  int len = strlen (name);

  new->next = vars;
  new->name = xmalloc (len + 1);
  memcpy (new->name, name, len + 1);
  new->value = value;

  DODEBUG (2, printf ("Adding variable `%s'\n", new->name));

  vars = new;
  return 0;
}

/* See `parse.h' for a description.  */
char *
get_variable (char *name)
{
  struct variable *ptr;

  DODEBUG (3, printf ("Entered %s (name: %s)\n", __FUNCTION__, name));

  for (ptr = vars; ptr != NULL; ptr = ptr->next)
    if (! strcmp (ptr->name, name))
      return ptr->value;

  panic ("Unknown variable: `%s'.\n", name);
  return NULL; /* Not reached.  */
}



/* Pass 1 and 2.  */
static struct command_line *
var_handling (int pass, int line_cnt, struct command_line *lines,
	      int callbacks_cnt, struct callback *callbacks)
{
  int line_no; /* Current line */
  struct command_line *result;

  /* Call the callback FNAME.  */
  char *function_call (char *fname)
    {
      int i;

      DODEBUG (3, printf ("Entered %s (fname: %s)\n", __FUNCTION__, fname));

      for (i = 0; i < callbacks_cnt; i ++)
	if (! strcmp (callbacks[i].name, fname))
	  return callbacks[i].callback (lines[line_no].handle);

      panic ("Unknown function: `%s'\n", fname);
      return NULL; /* Not reached.  */
    }

  DODEBUG (3, printf ("Entered %s (pass: %d, ...)\n", __FUNCTION__, pass));
  assert (pass == 1 || pass == 2);

  result = xmalloc (sizeof (struct command_line) * line_cnt);
  for (line_no = 0; line_no < line_cnt; line_no++)
    {
      char *input;
      int buf_size;
      char *buf;
      char *buf2;
      char *output;

      /* Add LEN bytes from DATA to the output line buffer.  */
      void add_output (char *data, int len)
	{
	  int offset = output - buf;

	  DODEBUG (3, do {
	    int i;
	    
	    printf ("Entering %s (...)\n"
		    "  buf_size: %d, offset: %d, len: %d; data: `",
		    __FUNCTION__, buf_size, offset, len);
	    for (i = 0; i < len; i ++)
	      printf ("%c", data[i]);
	    printf ("'\n");
	  } while (0));

	  if (offset + len > buf_size)
	    {
	      char *tmp;

	      buf_size += len; /* Usually slightly more than we need now.  */
	      DODEBUG (3, printf ("Resizing buffer to %d bytes\n", buf_size));
	      tmp = xrealloc (buf, buf_size);
	      if (tmp != buf)
		{
		  buf = tmp;
		  output = buf + offset;
		}
	    }
	  memcpy (output, data, len);
	  output += len;
	}

      /* We might have to increase the size of the buffer later, but
	 this is a reasonable default.  */
      buf_size = lines[line_no].length + 1;
      buf = xmalloc (buf_size);
      output = buf;

      /* Although we do not modify INPUT, it may not be NUL
	 terminated.  We copy it to make life eaiser.  */
      buf2 = input = malloc (buf_size);
      memcpy (input, lines[line_no].line, buf_size - 1);
      input[buf_size - 1] = '\0';

      DODEBUG (3, printf ("%s (pass = %d): parsing ``%s''\n",
			  __FUNCTION__, pass, input));

      while (1)
	{
	  char *found = strstr (input, (pass == 1) ? "$(" : "${");

	  DODEBUG (3, printf ("found = %s\n", found));
	  if (found == NULL)
	    break;

	  /* Check for `\$', if not at the beginning of the line.  */
	  if (found != input && *(found - 1) == '\\')
	    {
	      add_output (input, found - input + 2);
	      input = found + 2;
	      continue;
	    }

	  if (found != input)
	    add_output (input, found - input);

	  input = found + 2; /* Skip `$(' or `${'.  */

	  if (pass == 1)
	    {
	      char *var;
	      char *end;
	      char *value;

	      /* Create the variable.  */
	      found = strchr (input, '=');
	      end = strchr (input, ')');

	      if (! end)
		panic ("Missing `)' in line %d\n", line_no + 1);

	      *end = '\0';

	      if (found > end)
		found = 0;

	      if (found)
		{
		  *found = '\0';
		  var = input;
		  input = found + 1;
		}
	      else
		var = 0;

	      value = function_call (input);
	      assert (value);
	      if (var)
		assert (add_variable (var, value) == 0);
	      else
		free (value);

	      input = end + 1;
	    }
	  else
	    {
	      char *v;
	      found = strchr (input, '}');
	      if (found == NULL)
		panic ("Missing `}' in line %d\n", line_no + 1);
	      *found = '\0';
	      v = get_variable (input);
	      add_output (v, strlen (v));
	      input = found + 1;
	    }
	}
      add_output (input, strlen (input) + 1);
      result[line_no].line = buf;
      result[line_no].length = strlen (buf) + 1;
      result[line_no].handle = lines[line_no].handle;
      DODEBUG (2, printf ("Pass %d - output line: ``%s''\n", pass, buf));

      free (buf2);
    }
  return result;
}

struct argv *
split_into_argvs (int line_cnt, struct command_line *lines)
{
  struct argv *result;
  int line_no;

  DODEBUG (3, printf ("Entered %s (...)\n", __FUNCTION__));
  result = xmalloc (sizeof (struct argv) * line_cnt);
  for (line_no = 0; line_no < line_cnt; line_no++)
    {
      DODEBUG (3, printf ("%s: Splitting line %d\n",
			  __FUNCTION__, line_no));

      result[line_no].argc = 1;
      result[line_no].argz =
	({
	  char *in = lines[line_no].line;
	  int left = lines[line_no].length;
	  char *buf = xmalloc (lines[line_no].length + 2);
	  char *out;
	  int in_quotes = 0;
	  int seen_space = 1; /* Ignore whitespace at the beginning of a
				 line.  */

	  for (out = buf; *in != '\0' && left > 0; in++, left --)
	    switch (*in)
	      {
	      case '"':
		in_quotes = !in_quotes;
		seen_space = 0;
		break;

	      case ' ':
	      case '\t':
		if (in_quotes)
		  {
		    *out = *in;
		    out ++;
		  }
		else if (!seen_space)
		  {
		    *out = '\0';
		    out ++;
		    result[line_no].argc++;
		    seen_space = 1;
		  }
		break;

	      case '\\':
		in++;
		left --;
		if (left == 0)
		  panic ("Incomplete escape sequence at the end of line %d.",
			 line_no + 1);

		switch (*in)
		  {
		  case 'n': *out = '\n'; break;
		  case 't': *out = '\t'; break;
		  case '"': *out = '"'; break;
		  case ' ': *out = ' '; break;
		  case '$': *out = '$'; break;
		  case '\0':
		    /* We might want to continue reading in the next
		       line in this case, but we should not forget to
		       realloc our output buffer then.  */
		    panic ("Incomplete escape sequence at the end of line %d.",
			   line_no + 1);
		  default:
		    panic ("Unknown escape sequence `\\%c' in line %d\n",
			   *in, line_no + 1);
		  }
		out ++;
		seen_space = 0;
		break;

	      default:
		*out = *in;
		out ++;
		seen_space = 0;
		break;
	      }

	  if (in_quotes)
	    panic ("Unmatched `\"' in line %d\n", line_no + 1);

	  *out = '\0';
	  if (seen_space)
	    result[line_no].argc--;
	  else
	    *(++out) = '\0';
	  result[line_no].size = out - buf + 1;
	  buf; /* Return value.  */
	});

      DODEBUG (3, do {
	int i;
	char *s;

	printf ("argz:\n");
	for (i = 0, s = result[line_no].argz;
	     i < result[line_no].argc;
	     i ++, s += strlen (s) + 1)
	  printf ("\t%s\n", s);
      } while (0));
    }
  return result;
}



/* See `parse.h' for a description.  */
struct argv *
parse_command_lines (struct command_line *lines, int line_cnt,
		     struct callback *callbacks, int callback_cnt)
{
  struct command_line *after_first_pass;
  struct command_line *after_second_pass;
  struct argv *result;

  /* Free each element in STRINGS, which is an array terminated by an
     element containing a null pointer, as well as the array itself.  */
  inline void
    free_lines (struct command_line *lines)
    {
      struct command_line *ptr;
      int i = line_cnt;

      for (ptr = lines; i != 0; ptr++, i--)
	free (ptr->line);
      free (lines);
    }

  DODEBUG (3, printf ("Entered %s (%d lines, %d callbacks\n",
		      __FUNCTION__, line_cnt, callback_cnt));

  after_first_pass = var_handling (1, line_cnt, lines, callback_cnt,
				   callbacks);

  after_second_pass = var_handling (2, line_cnt, after_first_pass,
				    0, NULL);
  free_lines (after_first_pass);
  /* Do not free the variables here, because our caller might want to
     use them later.  */

  result = split_into_argvs (line_cnt, after_second_pass);
  free_lines (after_second_pass);

  return result;
}



/******************** Test code ********************/

#ifdef  TEST_PARSER
static char *
cb (void *handle)
{
  char *str = handle;
  char *new = xmalloc (strlen (str) + 1);

  DODEBUG (2, printf ("Callback with arg `%s'\n", str));
  memcpy (new, str, strlen (str) + 1);
  return new;
}

static struct callback callbacks[] =
{
  {"foo", cb},
  {"bar", cb},
  {0}
};

static struct command_line *
strs2cmds (char **argv, int argc)
{
  struct command_line *result = xmalloc (sizeof (struct command_line) * argc);

  while (argc--)
    {
      result[argc].line = argc[argv];
      result[argc].handle = (argc % 2) ? "<foo>" : "<bar>";
    }
  return result;
}

#include <argz.h> /* For argz_stringify () */

int
main (int argc, char **argv)
{
  struct argv *result = parse_command_lines (strs2cmds (++argv, argc - 1),
					     argc - 1, callbacks);

  DODEBUG (2, do {
    printf ("------------------\n");
    while (--argc)
      {
	argz_stringify (result->argz, result->size, '|');
	printf ("argc: %d, size: %d, argz: `%s'\n",
		result->argc, result->size, result->argz);
	result++;
      }
  } while (0));

  return 0;
}
#endif
