#include <stdio.h>
#include <stdlib.h>
#include <elf.h>
#include <assert.h>
#include <string.h>
#include <machine/vmparam.h>

#include <hurd/ihash.h>
#include <hurd/debug.h>
#include <sys/panic.h>

#include <boot-modules.h>

#include <l4/l4.h>
#include <l4io.h>
#include <l4/sigma0.h>

#include "elf-load.h"
#include "parse.h"
#include "task.h"
#include "task_client.h"

L4_Fpage_t kip_area, utcb_area;
L4_ThreadId_t master_tid, task_server_tid, pager_tid;

/* The Kernel Interface page.  */
L4_KernelInterfacePage_t *kip;

static ihash_t vpage_hash;

static L4_Word_t
find_page (task_t task, L4_Word_t vaddress)
{
  L4_Word_t p;

  if (vpage_hash)
    p = (L4_Word_t) ihash_find (vpage_hash, task | vaddress);
  else
    p = 0;

  DODEBUG (3, printf ("find_page (%x, %x) => %x\n",
		      task, vaddress, p));
  return p;
}

/* Task is the owner; vaddress is where the page should be installed.
   Pages owned by task "0" are available to all.  */
static void
hash_page (task_t task, L4_Word_t vaddress, void *paddress)
{
  L4_Word_t paddr;

  DODEBUG (3, printf ("%s: Hashing %x:%x (%x) to %x\n",
		      __FUNCTION__, vaddress, task, vaddress | task,
		      (int) paddress));

  if (task & PAGE_MASK)
    panic ("Task id (%x) too large to hash.", task);
  assert ((vaddress & ~PAGE_MASK) == 0);

  if ((paddr = find_page (task, vaddress)))
    panic ("%s: Already hashed %x:%x to %x; now trying to hash to %x\n",
	   __FUNCTION__, vaddress, task, paddr, (int) paddress);

  if (! vpage_hash)
    {
      assert (ihash_create (&vpage_hash) == 0);
      DODEBUG (3, printf ("Allocated vpage_hash = %p+(%x)\n",
			  vpage_hash, sizeof (vpage_hash)));
    }

  assert (ihash_add (vpage_hash, task | vaddress,
		     (void *) paddress, 0) == 0);
}

static void *
allocate_page (void)
{
  L4_Fpage_t p;

#define sigma0_tid() (L4_GlobalId (kip->ThreadInfo.X.UserBase, 1))
  p =  L4_Sigma0_GetPage (sigma0_tid (),
			  L4_FpageLog2 (-1UL << 10,
					PAGE_SHIFT));
  p.raw &= ~0x3ff;

  DODEBUG (2, printf ("%s: Allocated page %x\n",
		      __FUNCTION__, p.raw));

  return (void *) p.raw;
}

void
pager_server (void)
{
  L4_ThreadId_t tid;
  L4_MsgTag_t tag;
  L4_Msg_t msg;
  L4_Word_t addr;
  L4_MapItem_t page;

  DODEBUG (0, printf ("%s: Pager is up.\n", __FUNCTION__));
    
  for (;;)
    {
      tag = L4_Wait (&tid);
	
      for (;;)
	{
	  L4_StoreMsg (tag, &msg);

	  if (L4_IpcFailed (tag))
	    {
	      enter_kdebug ("Receive failed.");
	      break;
	    }
#ifndef HAVE_PROPAGATION
	  else if (L4_UntypedWords (tag) == 3 && L4_TypedWords (tag) == 0)
	    //		   && L4_SameThreads (tid, master_tid))
	    /* Start a task.  */
	    {
	      L4_Word_t ip, sp;
	      L4_Get_Word (&msg, 0, &ip);
	      L4_Get_Word (&msg, 1, &sp);
	      L4_Get_Word (&msg, 2, (L4_Word_t *) &tid);
	      DODEBUG (1, printf ("%s: Starting task %x. ip = %x; "
				  "sp = %x\n",
				  __FUNCTION__, * (int *) &tid, ip, sp));
	      L4_Clear (&msg);
	      L4_Append_Word (&msg, ip);
	      L4_Append_Word (&msg, sp);
	    }
#endif
	  else if (L4_UntypedWords (tag) == 2 && L4_TypedWords (tag) == 0)
	    {
	      L4_Word_t paddr;
	      task_t taskid;

	      L4_Clear (&msg);
	      L4_Get_Word (&msg, 0, &addr);
	      DODEBUG (2, printf ("%s: Page fault by %x at %x\n",
				  __FUNCTION__,
				  * (int *) &tid, addr));
	      addr &= PAGE_MASK;

	      taskid = L4_Version (tid);
	      paddr = find_page (taskid, addr);
	      if (! paddr)
		paddr = find_page (0, addr);
	      if (! paddr)
		{
		  DODEBUG (0, printf ("%s: Invalid page fault by "
				      "%x:%x at %x\n",
				      __FUNCTION__,
				      * (int *) &tid, taskid, addr));
		  break;
		}

	      DODEBUG (2, printf ("%s: Providing %x to service fault "
				  "by %x at %x",
				  __FUNCTION__, paddr, * (int *) &tid,
				  addr));

	      page = L4_MapItem (L4_Fpage_Set_Attrs (L4_FpageLog2 (paddr,
								   PAGE_SHIFT),
						     L4_FullyAccessible),
				 addr);
	      L4_Append_MapItem (&msg, page);
	    }
	  else
	    {
	      panic ("Unknown message from %x", * (int *) &tid);
	      break;
	    }

	  L4_LoadMsg (&msg);
	  tag = L4_ReplyWait (tid, &tid);
	}
    }
}

static void
send_startup_ipc (L4_ThreadId_t id, L4_Word_t ip, int argc, char *argz,
		  int argz_size)
{
  L4_Word_t *sp = allocate_page ();
  void *argz_copy;
  L4_Msg_t msg;

  hash_page (L4_Version (id), (L4_Word_t) sp, (void *) sp);

  /* The stack layout.

     On the top of the stack, we place an argz (being careful to make
     sure that the start it is aligned on a word boundary) containing
     the arguments (i.e. the command line).  This is followed by the
     size of the argz (in bytes).  The exact format of an argz data
     structure is described in the glibc manual.  After this, the
     thread id of the task server and the system pager are placed on
     the stack.  Lastly, a zero is appended.  This final value exists
     only for added convenience as normally: the stack pointer points
     to the return address; if the start function tries to return, it
     will segfault; and it means we can typically (but certainly not
     always) avoid an assembly code entry point.

          +--------------+  /  Space may be empty.  Must round the argz size up to
          +--------------+  \  the next aligned word to find the task server id.
          |              |
          |     argz     |
          |              |
          +--------------+
          |  argz size   |
          +--------------+
          |  task server |
          +--------------+
          | system pager |
          +--------------+
          |      0       |
          +--------------+ <-- stack pointer
   */

  sp += PAGE_SIZE / sizeof (L4_Word_t);

  /* Setting up the stack is mostly unportable: first of all, we
     assume that the stack grows down.  */

  /* We only allocate a page and we need some room to work before we
     switch stacks.  */
  assert (argz_size < PAGE_SIZE - 100);

  /* Copy the argz.  */
  sp -= (argz_size + sizeof (L4_Word_t) - 1) / sizeof (L4_Word_t);
  argz_copy = sp;
  memcpy (argz_copy, argz, argz_size);

  *(-- sp) = argz_size;

  /* The system pager and the task server thread ids.  */
  *(-- sp) = * (int *) &task_server_tid;
  *(-- sp) = * (int *) &pager_tid;

  /* Skip a byte; normally this would be the return address.  */
  *( --sp) = 0;

  DODEBUG (1, printf ("%s: Sending startup message to %x, "
		      "(ip=%x, sp=%p)\n",
		      __FUNCTION__, * (int *) &id, ip, sp));

  L4_Clear (&msg);
#ifdef HAVE_PROPAGATION
  L4_Set_VirtualSender (pager_tid);
  L4_Set_Propagation (&msg.tag);
#endif
  L4_Append_Word (&msg, ip);
  L4_Append_Word (&msg, (L4_Word_t) sp);
#ifndef HAVE_PROPAGATION
  L4_Append_Word (&msg, *(L4_Word_t *) &id);
  id = pager_tid;
  DODEBUG (2, printf ("%s: Redirecting start request to pager (%x).\n",
		      __FUNCTION__, * (int *) &id));
#endif
  L4_LoadMsg (&msg);
  L4_Send (id);
}

struct handle
{
  task_t tid;
  void *data;
};

/* Read in the meta data of the executable.  */
static int
exec_read (void *handle, vm_offset_t file_ofs,
	   void *buf, vm_size_t size,
	   vm_size_t *out_actual)
{
  struct handle *h = handle;

  memcpy (buf, h->data + file_ofs, size);
  *out_actual = size;
  return 0;
}

/* "Read" in parts of the executable.  As we map virtual to physical
   memory one-to-one, this is trivial implement, however, limited in
   functionality: any overlap kills us as we cannot remap.  */
static int
exec_read_exec (void *handle,
	        vm_offset_t file_ofs, vm_size_t file_size,
	        vm_offset_t mem_addr, vm_size_t mem_size,
	        exec_sectype_t section_type)
{
  struct handle *h = handle;
  vm_offset_t start = mem_addr;
  int offset;
  int todo;
  void *p;

  DODEBUG (4, printf ("%s (file_ofs = %d:%x, file_size = %d:%x, "
		      "mem_addr = %d:%x, mem_size = %d:%x\n",
		      __FUNCTION__, file_ofs, file_ofs,
		      file_size, file_size, mem_addr, mem_addr,
		      mem_size, mem_size));

  /* We do not need debugging sections.  */
  if (! (section_type & (EXEC_SECTYPE_ALLOC)))
    return 0;

  mem_size -= file_size;

  for (offset = start & ~PAGE_MASK,
	 todo = (file_size > PAGE_SIZE - offset
		   ? PAGE_SIZE - offset
		   : file_size);
       file_size > 0;
       offset = 0, start += todo, file_ofs += todo,
	 file_size -= todo,
	 todo = (file_size > PAGE_SIZE - offset
		   ? PAGE_SIZE - offset
		   : file_size))
    {
      p = allocate_page ();
      assert (p);
      memcpy (p + offset, h->data + file_ofs, todo);
      DODEBUG (4, printf ("%s: memcpy (%x + %d(%x), file + %d(%x), "
			  "%d(%x)); \nmemsize = %d(%x) "
			  "filesize = %d(%x) start = %d(%x)\n",
			  __FUNCTION__, (L4_Word_t) p, offset, offset,
			  file_ofs, file_ofs, todo, todo,
			  mem_size, mem_size, file_size, file_size,
			  start, start));

      hash_page (h->tid, start & PAGE_MASK, p);
    }

  assert (file_size == 0);

  offset = start & ~PAGE_MASK;
  if (offset)
    {
      memset (p + offset, 0, PAGE_SIZE - offset);
      DODEBUG (4, printf ("%s: memset (%x + %d(%x), 0, "
			  "PAGE_SIZE - %d(%x)); "
			  "memsize = %d(%x) filesize = %d(%x)\n",
			  __FUNCTION__, (L4_Word_t) p, offset, offset,
			  offset, offset, mem_size, mem_size,
			  file_size, file_size));
      start += PAGE_SIZE - offset;
    }

  assert ((start & ~PAGE_MASK) == 0);

  for (todo = mem_size > PAGE_SIZE ? PAGE_SIZE : mem_size;
       mem_size > 0;
       mem_size -= todo, start += PAGE_SIZE,
	 todo = mem_size > PAGE_SIZE ? PAGE_SIZE : mem_size)
    {
      p = allocate_page ();
      assert (p);
      memset (p, 0, todo);
      DODEBUG (4, printf ("%s: memset (%x, 0, %d(%x)); "
			  "mem_size = %d(%x); start = %d(%x)\n",
			  __FUNCTION__, (L4_Word_t) p, todo, todo,
			  mem_size, mem_size, start, start));

      hash_page (h->tid, start & PAGE_MASK, p);
    }

  return 0;
}

static L4_ThreadId_t
start_server (char *server, L4_Word_t func)

{
  L4_ThreadId_t tid = L4_GlobalId (++ last_threadno,
				   L4_Version (L4_Myself ()));
  /* XXX: Assuming that the stack grows down.  */
  L4_Word_t stack = (L4_Word_t) ((void *) allocate_page () + PAGE_SIZE);

  L4_ThreadControl (tid, L4_Myself (), L4_Myself (),
		    L4_nilthread);
  L4_Start_SpIp (tid, stack, func);

  DODEBUG (1, printf ("%s: started %s: stack = %x, tid = %x\n",
		      __FUNCTION__, server, stack, * (int *) &tid));
  return tid;
}

extern void task_server (void);

int
main (void)
{
  kip = L4_GetKernelInterface ();

#define UTCB_SIZE (16 * 1024 * 1024)
#define UTCB_START ((VM_MAX_ADDRESS - UTCB_SIZE) + 1)
  utcb_area = L4_Fpage (UTCB_START, UTCB_SIZE);

  kip_area = L4_FpageLog2 ((L4_Word_t) (UTCB_START
					- (1 << kip->KipAreaInfo.X.s)),
			   kip->KipAreaInfo.X.s);

  DODEBUG (1, printf ("utcb_area = %x; kip_area = %x\n",
		      utcb_area.raw, kip_area.raw));

  DODEBUG (0, printf ("Hello, World!\nroot-server: " "tid = %d\n",
		      L4_Myself ().raw));

  /* Fault in all of our pages as otherwise our pager might wedge.  */
  {
    extern L4_Word_t _start, _end;
    void *p;

    for (p = &_start; (L4_Word_t) p < (L4_Word_t) &_end; p += PAGE_SIZE)
      *(volatile int *) p;
  }

  /* Start the servers.  */

  master_tid = L4_Myself ();
  last_threadno = L4_ThreadNo (L4_Myself ());

  pager_tid = start_server ("pager",
			    (L4_Word_t) pager_server);
  task_server_tid = start_server ("task",
				  (L4_Word_t) task_server);

  /* Load the os personality.  */

  {
    int i;
    struct boot_modules_header *bmh = (void *) kip->BootInfo;
    struct boot_module *m;
    struct command_line command_lines[bmh->module_count];
    struct parse_handle
    {
      struct boot_module *m;
      thread_t threadid;
      L4_Word_t ip;
      int do_start;
    } parse_handles[bmh->module_count];
    struct argv *argvs;

    char *do_task_create (void *x)
      {
	struct parse_handle *ph = x;
	error_t err;
	struct handle h;
	task_t tid;
	CORBA_Environment env = idl4_default_environment;
	exec_info_t exec_info;
	L4_ThreadId_t thread;
	char *s;

	printf ("Loading module\n");
	assert (! task_create (task_server_tid, 0, &tid, &env));
	h.data = (void *) ph->m->start;
	h.tid = tid;

	err = exec_load (exec_read, exec_read_exec, &h,
			 &exec_info);

	if (err)
	  panic (" failed: %s", ex_error_message (err));
	else
	  printf (" ok.\n");

	env = idl4_default_environment;
	assert (! thread_create (task_server_tid, tid,
				 * (int *) &pager_tid,
				 (int *) &thread, &env));

	ph->threadid = * (int *) &thread;
	ph->ip = exec_info.entry;

	s = malloc (10);
	assert (s);
	snprintf (s, 10, "%d", tid);
	return s;
      }

    char *do_task_resume (void *x)
      {
	struct parse_handle *ph = x;
	if (! ph->ip)
	  panic ("Task not created; cannot be started.");
	else
	  ph->do_start = 1;

	return strdup ("");
      }

    struct callback callbacks [] =
      {
	{
	  name: "task-create",
	  callback: do_task_create
	},
	{
	  name: "task-resume",
	  callback: do_task_resume
	}
      };

    /* Predefined variables.  */

    {
      char buf[10];

      buf[sizeof (buf) - 1] = 0;
      snprintf (buf, sizeof (buf) - 1, "%d", pager_tid);
      printf ("%s\n", buf);
      add_variable ("root-pager", strdup (buf));
      snprintf (buf, sizeof (buf) - 1, "%d", task_server_tid);
      printf ("%s\n", buf);
      add_variable ("task-server", strdup (buf));
    }

    /* Variables passed on the command line.  */

    {
      char *s;
      char *equal;

      s = malloc (bmh->commandline_length + 1);
      assert (s);
      memcpy (s, bmh->commandline, bmh->commandline_length);
      s[bmh->commandline_length] = 0;

      printf ("%s\n%d modules.\n", s, bmh->module_count);

      equal = s;
      while ((equal = strchr (equal, '=')))
	{
	  char *var;
	  char *value;

	  /* XXX Does not do quoting.  */
	  var = equal;
	  *equal = 0;
	  while (var >= s && *var != ' ' && *var != '\t')
	    var --;
	  if (*var == ' ' || *var == '\t')
	    var ++;

	  value = ++ equal;
	  while (*equal != '\0' && *equal != ' ' && *equal != '\t')
	    equal ++;
	  *equal = 0;
	  equal ++;

	  add_variable (var, strdup (value));
	}

      free (s);
    }

    /* Set up the servers.  */

    for (i = 0, m = (void *) bmh + bmh->modules_offset;
	 i < bmh->module_count;
	 i ++, m = (void *) m + m->size)
      {
	parse_handles[i].m = m;
	parse_handles[i].threadid = 0;
	parse_handles[i].do_start = 0;
	
	command_lines[i].line = m->commandline;
	command_lines[i].length = m->commandline_length;
	command_lines[i].handle = &parse_handles[i];
      }

    argvs = parse_command_lines (command_lines, bmh->module_count, callbacks,
				 sizeof (callbacks) / sizeof (*callbacks));

    DODEBUG (1, printf ("%s: Done parsing modules.\n", __FUNCTION__));

    parser_cleanup ();

    for (i = 0; i < bmh->module_count; i ++)
      if (parse_handles[i].do_start != 0)
	{
	  DODEBUG (1, printf ("Starting module %d (threadid = %x; ip = %x)\n",
			      i, * (int *) &parse_handles[i].threadid,
			      parse_handles[i].ip));

	  send_startup_ipc (* (L4_ThreadId_t *) &parse_handles[i].threadid,
			    parse_handles[i].ip, argvs[i].argc, argvs[i].argz,
			    argvs[i].size);
	}
  }

  /* XXX: Should kill ourselves, but oh yeah, that does not work!  */
  for (;;)
    L4_Yield ();

  /* Never returns.  */
  return 0;
}
