/* Task and thread interface.
   Copyright (C) 2002 Free Software Foundation, Inc.
   Written by Neal H. Walfield.

   This file is part of the GNU Hurd.

   The GNU Hurd is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   The GNU Hurd is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA. */

#include <assert.h>
#include <stdio.h>
#include <sys/malloc.h>
#include <sys/panic.h>
#include <hurd/ihash.h>
#include <hurd/debug.h>
#include <l4/l4.h>

#include <hurd/l4/interfaces.h>

#include "spin-lock.h"

#define CORBA_free free

#include "task.h"
#include "task_server.h"

/* Interface task */
int last_threadno;

/* The maximum number of tasks in the system.  We always store the
   task id number in the version field, for 32 bit systems, the
   version field is 14 bits wide.  */
#define LAST_TASKID ((1 << 14) - 1)
#define FIRST_TASKID (L4_Version (task_server_tid) + 1)

struct task
{
  /* Task id number; zero is invalid.  */
  task_t id;

  /* The creator, i.e. the parent.  This may be NULL if the task
     created itself (in the case of the root_server) or if the parent
     died.  */ 
  task_t parent;

  /* Threads in the task may only be scheduled if the suspend count is
     less than or equal to zero.  */
  int suspend_count;

  /* L4 does not have address space identifiers.  Instead, a thread
     identifier is used to denote an address space.  In L4, if all of
     the threads in an address space are destroyed, the address space
     is also destroyed; this is not what we want.  To work around
     this, we always preserve a single thread.  When an address space
     is created SPACE is set to the initial inactive thread identifier
     (thread_hash is left empty and THREAD_COUNT is set to 0).  When
     the first thread is create by the user, this thread will be mark
     active and entered into the hash.

     This is completely safe as an active thread may only be started
     in one of two ways: via the exchange registers and thread control
     system calls.  The former can only be used by a thread in the
     same address space.  As there are no active threads in a new
     address space, the thread cannot be made active.  Thread control
     may only be called by a privileged server.  Thus, thread creation
     must always go through us.

     As threads are created and destroyed, the initial thread may be
     terminated.  In this case, another thread must be choosen to
     identify the space--the choice is irrelevant just so long as the
     thread is valid.  When the last thread is about to be destroyed,
     we instead mark it as inactive, remove it from the hash and
     deallocate the hash.  This returns us to, essentially, the
     initial state.  */
  int thread_count;
  L4_ThreadId_t space;
  ihash_t thread_hash;

  /* ihash private data.  */
  void ** locp;
};

struct thread
{
  /* The L4 thread id.  */
  L4_ThreadId_t id;

  /* The thread by only be scheduled if both its suspend cout and that
     of the containing task is less than or equal to zero.  */
  int suspend_count;

  /* ihash private data.  */
  void ** locp;
};

/* Hashes a task id to a struct task *.  */
static ihash_t task_hash;

static spin_lock_t lock = SPIN_LOCK_INITIALIZER;

/* Returns the apparent task identifier given a thread id.  This must
   still be verified.  */
static task_t
threadid_to_taskid (L4_ThreadId_t thread)
{
  return L4_Version (thread);
}

static struct task *
find_task (task_t t)
{
  return ihash_find (task_hash, t);
}

static L4_ThreadId_t
alloc_threadno (task_t task)
{
  /* XXX: We need to do recycling.  */
  return L4_GlobalId (++ last_threadno, task);
}

static task_t
alloc_taskno (void)
{
  static int last_task_id;
  int tries = 0;

  do
    {
      last_task_id ++;
      if (last_task_id > LAST_TASKID) /* Size of the version field.  */
	{
	  last_task_id = FIRST_TASKID;
	  continue;
	}

      tries ++;
      if (tries == LAST_TASKID - FIRST_TASKID + 1)
	{
	  DODEBUG (0, panic ("No free task ids!"));
	  return 0;
	}
    }
  while (last_task_id % 64 == 0
	 || ihash_find (task_hash, last_task_id));

  return last_task_id;
}

/* Initialize the server.  */
void
task_server_init (void)
{
#define FUCK_NO_MEMORY() \
	panic ("No memory to boot system (%s:%d)", \
	       __FILE__, __LINE__)

  error_t err;
  struct task *task;
  int i;
  L4_ThreadId_t system_thread[] = { master_tid, task_server_tid, pager_tid };

  DODEBUG (1, printf ("Task server is up.\n"));

  /* We make a few assumptions about these sizes.  Be paranoid and
     check that they hold.  */
  assert (sizeof (task_t) == sizeof (void *));
  assert (sizeof (task_t) == sizeof (L4_ThreadId_t));

  err = ihash_create (&task_hash);
  if (err)
    FUCK_NO_MEMORY ();

  task = malloc (sizeof (struct task));
  if (! task)
    FUCK_NO_MEMORY ();

  err = ihash_add (task_hash, L4_Version (system_thread[0]),
		   task, &task->locp);
  if (err)
    FUCK_NO_MEMORY ();

  task->suspend_count = 0;
  task->space = system_thread[0];
  task->thread_count = 0;

  err = ihash_create (&task->thread_hash);
  if (err)
    FUCK_NO_MEMORY ();

  for (i = 0; i < sizeof (system_thread) / sizeof (*system_thread); i ++)
    {
      struct thread *thread = malloc (sizeof (struct thread));
      if (! thread)
	FUCK_NO_MEMORY ();

      task->thread_count ++;
      thread->id = system_thread[i];
      thread->suspend_count = 0;

      err = ihash_add (task->thread_hash, * (int *) &thread->id,
		       thread, &thread->locp);
      if (err)
	FUCK_NO_MEMORY ();
    }
#undef FUCK_NO_MEMORY
}

IDL4_INLINE error_t
task_create_implementation (CORBA_Object _caller,
			    const task_t redirector,
			    task_t *tid,
			    idl4_server_environment *_env)

{
  error_t err;
  struct task *task;
  L4_ThreadId_t threadid;

  DODEBUG (3, printf ("Entered %s.  caller = %x; redirector = %x\n",
		      __FUNCTION__, * (int *) &_caller, redirector));

  /* XXX: Anyone can create tasks; we need some policy.  */
  if (redirector && ! find_task (redirector))
    {
      err = EPERM;
      goto out;
    }

  spin_lock (&lock);

  task = malloc (sizeof (struct task));
  if (! task)
    {
      DODEBUG (0, panic ("Failed to allocate a task structure."));
      err = ENOMEM;
      goto out_unlock;
    }

  *tid = task->id = alloc_taskno ();
  DODEBUG (1, printf ("Allocated tid: %x\n", task->id));

  err = ihash_add (task_hash, task->id, task, &task->locp);
  if (err)
    {
      DODEBUG (0, printf ("ihash_add failed"));
      goto out_dealloc;
    }

  threadid = alloc_threadno (task->id);
  DODEBUG (1, printf ("Allocated threadid: %x for task %x\n",
		      threadid.raw, task->id));

  task->parent = threadid_to_taskid (_caller);
  assert (ihash_find (task_hash, task->parent));

  task->suspend_count = 0;
  task->space = threadid;
  task->thread_count = 0;
  task->thread_hash = 0;

  err = ! L4_ThreadControl (threadid, threadid, L4_Myself (),
			    L4_nilthread);
  if (err)
    {
      DODEBUG (0, panic ("Error creating new task %x\n", threadid.raw));
      err = ENOMEM;
      goto out_dealloc;
    }

  {
    L4_Word_t control;
    err = ! L4_SpaceControl (threadid, 0, kip_area, utcb_area,
			     L4_anythread, &control);
    if (err)
      {
	DODEBUG (0, panic ("Error setting up address space for "
			   "task %x\n", threadid.raw));
	err = EGRATUITOUS;
	goto out_dealloc;
      }
  }

  DODEBUG (4, printf ("task = %p\n\t->id = %x\n\t->parent = %x\n\t"
		      "->suspend_count = %x\n\t->thread_count = %d\n\t"
		      "->space = %x\n\t->thread_hash = %p\n",
		      task, task->id, task->parent, task->suspend_count,
		      task->thread_count, * (int *) &task->space,
		      task->thread_hash));

 out_dealloc:
  if (err)
    free (task);

 out_unlock:
  spin_unlock (&lock);

 out:
  DODEBUG (3, printf ("Leaving %s.  caller = %x\n",
		      __FUNCTION__, * (int *) &_caller));
  return err;
}

IDL4_PUBLISH_TASK_CREATE (task_create_implementation);

IDL4_INLINE error_t
task_terminate_implementation (CORBA_Object _caller, const task_t taskid,
			       idl4_server_environment * _env)
{
  error_t err = 0;
  struct task *t;

  error_t
    terminate (void *x)
    {
      struct thread *thread = x;

      DODEBUG (1, printf ("Task %x terminating thread %x.\n",
			  taskid, * (int *) &thread->id));

      assert (L4_ThreadControl (thread->id,
				L4_nilthread,
				L4_nilthread,
				L4_nilthread) == 1);
      t->thread_count --;

      if (L4_SameThreads (thread->id, _caller))
	idl4_set_no_response (_env);

      /* No need to remove it from the hash table: the whole shebang
	 is going to be destroyed in a moment.  */
      free (thread);

      return 0;
    }

  DODEBUG (3, printf ("Entered %s.  caller = %x; taskid = %x\n",
		      __FUNCTION__, * (int *) &_caller, taskid));

  spin_lock (&lock);

  t = ihash_find (task_hash, taskid);
  if (! t)
    {
      DODEBUG (1, printf ("Invalid taskid %x", taskid));
      err = EINVAL;
      goto out;
    }

  DODEBUG (1, printf ("Task %x has %d threads.\n", taskid,
		      t->thread_count));
  assert (t->thread_count >= 0);
  if (t->thread_count > 0)
    ihash_iterate (t->thread_hash, terminate);
  assert (t->thread_count == 0);

  ihash_free (t->thread_hash);
  ihash_locp_remove (task_hash, t->locp);
  free (t);

 out:
  spin_unlock (&lock);

  DODEBUG (3, printf ("Leaving %s.  caller = %x\n",
		      __FUNCTION__, * (int *) &_caller));

  return err;
}

IDL4_PUBLISH_TASK_TERMINATE (task_terminate_implementation);

IDL4_INLINE error_t
task_resume_implementation (CORBA_Object _caller, const task_t taskid,
			    idl4_server_environment * _env)
{
  error_t err = 0;

  panic (__FUNCTION__ ": implement me");

  return err;
}

IDL4_PUBLISH_TASK_RESUME (task_resume_implementation);

IDL4_INLINE error_t
task_suspend_implementation (CORBA_Object _caller, const task_t taskid,
			     idl4_server_environment * _env)
{
  error_t err = 0;

  panic (__FUNCTION__ ": implement me");

  return err;
}

IDL4_PUBLISH_TASK_SUSPEND (task_suspend_implementation);

IDL4_INLINE error_t
task_priority_implementation (CORBA_Object _caller, const task_t task,
                              const CORBA_long priority,
                              const CORBA_boolean change_threads,
                              idl4_server_environment * _env)
{
  error_t __retval = 0;

  panic (__FUNCTION__ ": implement me");

  return __retval;
}

IDL4_PUBLISH_TASK_PRIORITY (task_priority_implementation);

void
task_discard (void)
{
}

void *task_vtable[TASK_DEFAULT_VTABLE_SIZE] = TASK_DEFAULT_VTABLE;

/* Interface thread */

IDL4_INLINE error_t
thread_create_implementation (CORBA_Object _caller,
			      const task_t taskid,
			      const thread_t pager,
			      thread_t *threadid,
			      idl4_server_environment * _env)
{
  error_t err;

  struct task *task;
  struct thread *thread;

  DODEBUG (3, printf ("Entered %s.  caller = %x; taskid = %d; pager = %x\n",
		      __FUNCTION__, * (int *) &_caller, taskid, pager));

  spin_lock (&lock);

  task = ihash_find (task_hash, taskid);
  if (! task)
    {
      DODEBUG (1, panic ("Taskid %x unknown!", taskid));
      err = EINVAL;
      goto out;
    }

  DODEBUG (4, printf ("task = %p\n\t->id = %x\n\t->parent = %x\n\t"
		      "->suspend_count = %x\n\t->thread_count = %d\n\t"
		      "->space = %x\n\t->thread_hash = %p\n",
		      task, task->id, task->parent, task->suspend_count,
		      task->thread_count, * (int *) &task->space,
		      task->thread_hash));

  thread = malloc (sizeof (struct thread));
  if (! thread)
    {
      DODEBUG (1, panic ("malloc failed!"));
      err = ENOMEM;
      goto out;
    }

  if (task->thread_count == 0)
    {
      thread->id = task->space;

      assert (task->thread_hash == 0);
      err = ihash_create (&task->thread_hash);
      if (err)
	{
	  DODEBUG (1, panic ("ihash_create failed"));
	  goto free_thread;
	}
    }
  else
    thread->id = alloc_threadno (task->id);

  *threadid = * (int *) &thread->id;
  task->thread_count ++;

  DODEBUG (1, printf ("%s: Using thread id = %x\n",
		      __FUNCTION__, * (int *) &thread->id));
  thread->suspend_count = 1;

  err = ihash_add (task->thread_hash, * (int *) &thread->id,
		   thread, &thread->locp);
  if (err)
    {
      DODEBUG (0, panic ("ihash_add failed"));
      goto free_thread;
    }

  err = ! L4_ThreadControl (thread->id, task->space, L4_Myself (),
			    * (L4_ThreadId_t *) &pager);
  if (err)
    {
      err = EGRATUITOUS;
      DODEBUG (0, panic ("Failed to start thread"));
      goto free_thread;
    }

 out:
  DODEBUG (3, printf ("Leaving %s.  caller = %x\n",
		      __FUNCTION__, * (int *) &_caller));

  spin_unlock (&lock);
  return err;

 free_thread:
  free (thread);
  goto out;
}

IDL4_PUBLISH_THREAD_CREATE (thread_create_implementation);

IDL4_INLINE error_t
thread_terminate_implementation (CORBA_Object _caller,
				 const thread_t threadid,
				 idl4_server_environment * _env)
{
  error_t err = 0;
  struct task *task;
  struct thread *thread;

  error_t
    find_spaceid (void *x)
    {
      struct thread *thread = x;

      task->space = thread->id;
      return 42;
    }

  DODEBUG (3, printf ("Entered %s.  caller = %x; threadid = %x\n",
		      __FUNCTION__, * (int *) &_caller, threadid));

  spin_lock (&lock);

  task = ihash_find (task_hash,
		     threadid_to_taskid (* (L4_ThreadId_t *) &threadid));
  if (! task)
    {
      DODEBUG (1, printf ("%s: Invalid threadid (no such task) %x\n",
			  __FUNCTION__, threadid));
      err = EINVAL;
      goto out;
    }

  thread = ihash_find (task->thread_hash, threadid);
  if (! thread)
    {
      DODEBUG (1, printf ("%s: Invalid threadid %x\n",
			  __FUNCTION__, threadid));
      err = EINVAL;
      goto out;
    }

  ihash_locp_remove (task->thread_hash, thread->locp);

  assert (task->thread_count > 0);
  if (task->thread_count == 1)
    {
      L4_ThreadId_t new_thread;

      assert (L4_SameThreads (task->space,
			      * (L4_ThreadId_t *) &threadid));

      /* There is no way to put a thread back into its initial
	 inactive state.  As at least one thread must exist in an
	 address space, we first create a new inactive thread and then
	 reap the target thread.  */

      new_thread = alloc_threadno (task->id);
      assert (! L4_ThreadControl (new_thread, task->space,
				  L4_Myself (), L4_nilthread));

      task->space = new_thread;

      ihash_free (task->thread_hash);
      task->thread_hash = 0;
    }

    assert (L4_ThreadControl (thread->id,
			      L4_nilthread,
			      L4_nilthread,
			      L4_nilthread) == 1);

    if (L4_SameThreads (task->space, thread->id))
      {
	assert (task->thread_count > 1);

	DODEBUG (2, printf ("Task %x is removing thread %x which "
			    "we are using as our space id.\n",
			    * (int *) &_caller, task->id));
	assert (ihash_iterate (task->thread_hash, find_spaceid) == 42);
	DODEBUG (2, printf ("New space id is %x.\n",
			    * (int *) &task->space));
      }
      
  task->thread_count --;
  assert (task->thread_count >= 0);

  free (thread);

  if (L4_SameThreads (* (L4_ThreadId_t *) &threadid, _caller))
    idl4_set_no_response (_env);

 out:
  spin_unlock (&lock);

  DODEBUG (3, printf ("Leaving %s.  caller = %x\n",
		      __FUNCTION__, * (int *) &_caller));

  return err;
}

IDL4_PUBLISH_THREAD_TERMINATE (thread_terminate_implementation);

IDL4_INLINE error_t
thread_resume_implementation (CORBA_Object _caller, const thread_t threadid,
			      idl4_server_environment * _env)
{
  error_t err = 0;

  /* implementation of thread::resume */

  return err;
}

IDL4_PUBLISH_THREAD_RESUME (thread_resume_implementation);

IDL4_INLINE error_t
thread_suspend_implementation (CORBA_Object _caller, const thread_t threadid,
			       idl4_server_environment * _env)
{
  error_t err = 0;

  /* implementation of thread::suspend */

  return err;
}

IDL4_PUBLISH_THREAD_SUSPEND (thread_suspend_implementation);

void *thread_vtable[THREAD_DEFAULT_VTABLE_SIZE] = THREAD_DEFAULT_VTABLE;

void
task_server(void)
{
  idl4_msgbuf_t msgbuf;
  L4_MsgTag_t msgtag;
  L4_ThreadId_t partner;
  int cnt;

  task_server_init ();

  idl4_msgbuf_init (&msgbuf);
  for (cnt = 0; cnt < THREAD_STRBUF_SIZE + TASK_STRBUF_SIZE; cnt++)
    idl4_msgbuf_add_buffer (&msgbuf, malloc(8000), 8000);

  while (1)
    {
    restart:
      partner = IDL4_NILTHREAD;
      msgtag.raw = 0;
      cnt = 0;

      while (1)
        {
	  L4_Word_t id;

          idl4_reply_and_wait(&partner, &msgtag, &msgbuf, &cnt);
	  if (idl4_is_error (&msgtag))
	    break;

	  id = idl4_get_function_id (&msgtag);

	  switch (idl4_get_interface_id (&msgtag))
	    {
	    case TASK_INTERFACE_ID:
	      id &= TASK_FID_MASK;
	      idl4_process_request (&partner, &msgtag, &msgbuf, &cnt,
				   task_vtable[id]);
	      break;

	    case THREAD_INTERFACE_ID:
	      id &= THREAD_FID_MASK;
	      idl4_process_request (&partner, &msgtag, &msgbuf, &cnt,
				    thread_vtable[id]);
	      break;

	    default:
	      DODEBUG (0, printf ("%s: Dropping unknown message "
				  "(interface=%d; id = %d\n",
				  __FUNCTION__, id >> 6,
				  id & ((1 << 6) - 1)));
	      goto restart;
	    }
        }

      enter_kdebug("message error");
    }
}

